/********************************************************************
Copyright (c) 2001,2002, The Board of Trustees of the University of Illinois.
All Rights Reserved.

SABUL High Performance Data Transfer Protocol

Laboratory for Adavanced Computing (LAC)
National Center for Data Mining (NCDM)
University of Illinois at Chicago
http://www.lac.uic.edu/

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software (SABUL) and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the
following conditions:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimers.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimers in the documentation
and/or other materials provided with the distribution.

Neither the names of the University of Illinois, LAC/NCDM, nor the names
of its contributors may be used to endorse or promote products derived
from this Software without specific prior written permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE CONTIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
********************************************************************/


#include <iostream>
#include <unistd.h>
#include <stdlib.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


#include <sender.h>
#include <recver.h>

#include <popt.h>


#define USING_NONE		0	/* just an initial type, not really used */
#define USING_FILE		1	/* using file, with disk read/write */
#define USING_LENGTH		2	/* just send data in memory, not to use file, no disk w/r */
#define NO_SUCH_FILE		4	/* cannot find the file */


#define CMD_PORT		99999	/* the port number for this command line tool to exchange info */
#define FILELENGTH		100



#define DEFAULT_BLOCKSIZE 	(8*1024*1024)
#define DEFAULT_LENGTH 		(1024*DEFAULT_BLOCKSIZE)

/* definitions for popt to parse the argument */
#define RT_SABUL_VERSION	0x00000001
#define RT_VERBOSE		0x00000002
#define RT_SERVER_MODE		0x00000004
#define RT_CLIENT_MODE		0x00000008

#define RT_FILENAME		0x00000010
#define RT_OUTFILENAME		0x00000020
#define RT_TOTAL_LENGTH		0x00000040
#define RT_BLOCK_SIZE		0x00000080

#define RT_PACKET_SIZE		0x00000100


/* data struct for information exchange between command line tool server and client */
struct reqinfo
{
	char  type;	/*  USING_FILE, USING_LENGTH,  etc.  */	
	char  filename[FILELENGTH];
	long  length_h;
	long  length_l;   	/* total_length = length_h*1000000 + length_l  */
	
};




int main(int argc, char* argv[])
{

  int mode=-1;		/* set initial value to an invalid value */
  long long totaldatasize=DEFAULT_LENGTH;
  long blocksize = DEFAULT_BLOCKSIZE;

  int cli_port=CMD_PORT;	/* port no for this cmd line tools */
  struct sockaddr_in serveraddr;
  struct sockaddr_in clientaddr;

  reqinfo clientreq;
  clientreq.type=USING_NONE;
  
  bool block=false;    
  struct SabulOption sblop={7000, 7000, 1472, 1468, 365, 0, 4096000, 4096000, 1, 0};  
  int rate_init    = 300 ;		/* this is the default Initial rate in this program */
  int rate_l_limit = 0;
  int rate_u_limit = 1000;		/* upper limit is 1000Mbps, i.e. GbE */
  const char *package="SABUL";		
  double version=2.2;


  char *server_ipaddr = NULL;
  int  sabulport=-1;			/* set initial value to an invalid value */
  char *filename=NULL;
  char *outfilename=NULL;
  int fd;
  int v;
  
 
  int mysock0;		/* socket id to listen request */
  int mysock;		/* socket id to exchange information */
  char* data=NULL;
  timeval time1, time2;



    poptContext optCon;   /* context for parsing command-line options */

    struct poptOption optionsTable[] = {
  { "server",   's', POPT_ARG_NONE,   0,                  RT_SERVER_MODE, 
        "Run this program as a sabul server.", NULL },
  { "client", 'c', POPT_ARG_STRING, &server_ipaddr,   RT_CLIENT_MODE, 
        "Run this program as a sabul client, please specify the server address.", "ServerIP"},
  { "port",     'p', POPT_ARG_INT,   &sabulport,                  0, 
        "Run sabul on this port number", "7000" },
  { "packetsize",     'K', POPT_ARG_INT,   &sblop.m_iPktSize,         RT_PACKET_SIZE,
        "Specify the size of packets used in the transaction.", "1470" },
  { "sendbufsize",     'S', POPT_ARG_INT,   &sblop.m_lSendBufSize,         0,
        "Specify the size of UDP buffer on the sender side.", "65536" },
  { "recvbufsize",     'R', POPT_ARG_INT,   &sblop.m_lRecvBufSize,         0,
        "Specify the size of UDP buffer on the receiver side.", "65536" },
  { "initrate",     'I', POPT_ARG_INT,   &rate_init,         0,
        "Specify the initial throughput in Mbps.", "300" },
  { "ulimrate",     'U', POPT_ARG_INT,   &rate_u_limit,         0,
        "Specify the upper limit for the throughput in Mbps.", "1000" },
  { "llimrate",     'L', POPT_ARG_INT,   &rate_l_limit,         0,
        "Specify the lower limit for the throughput in Mbps.", "0" },
  { "totallength",     'l', POPT_ARG_LONG,   &totaldatasize,         RT_TOTAL_LENGTH, 
        "Specify the the size[UNIT: kilobytes] of data you want to transfer, data transferred only between memory, no disk read/write. Specify 0 if you want the program runs indefinitely.", "1000000" },
  { "blocksize",     'b', POPT_ARG_LONG,   &blocksize,         RT_BLOCK_SIZE,
        "Specify the the size of block you want to use during the transfering.", "1000000" },
  { "file",   'f', POPT_ARG_STRING, &filename,          RT_FILENAME, 
        "Specify the name of the file you want to get from the remote server.", "filename"},
  { "outputfile",   'o', POPT_ARG_STRING, &outfilename,          RT_OUTFILENAME, 
        "Specify the name of the file you want to use as the output.", "filename"},
  { "version",    'v', POPT_ARG_NONE,   0,                  RT_SABUL_VERSION, 
        "Display version information about this package.", NULL},
  { "verbose",    'V', POPT_ARG_INT,    &sblop.m_iVerbose,        RT_VERBOSE, 
        "Modify the level or mode of Verbosity in output messages, 0 means standard output, otherwise, write to log file.", "0"},
        POPT_AUTOHELP
        { NULL, 0, 0, NULL, 0 }
  };

    optCon = poptGetContext(NULL, argc, (const char **)argv, optionsTable, 0);
    //poptSetOtherOptionHelp(optCon, "<filename>");

    if ( argc < 2 )
    {
        poptPrintUsage(optCon, stderr, 0);
        poptPrintHelp(optCon, stderr, 0);
        exit(EXIT_FAILURE);
    }



    while ( (v=poptGetNextOpt(optCon)) >= 0)
    {
        switch(v)
        {
	    case RT_SERVER_MODE:
		mode=0;	   /* 0 means server mode */
		break;

	    case RT_CLIENT_MODE:
		mode=1;	  /* 1 means client mode */
		break;

	    case RT_FILENAME:
		if(clientreq.type!=USING_NONE)
		{
			fprintf(stderr, "Length parameter you specified is omitted, File will be used.\n");
		} 
		clientreq.type=USING_FILE;
		strncpy(clientreq.filename,filename, FILELENGTH);
		break;

	    case RT_OUTFILENAME:
		break;

	    case RT_TOTAL_LENGTH:
		if(clientreq.type!=USING_NONE)
		{
			fprintf(stderr, "Length parameter you specified is omitted, File will be used.\n");
		}
		else
		{ 
			totaldatasize *=1000;		/* the input value is in Kbytes */
		}

		clientreq.type=USING_LENGTH;
		if(totaldatasize==0)
		{
			totaldatasize = (long long) 1e60; 	/* set to fairly large value */
		}
		
		/* reorganize the value so that it can be transferred to the remote site *
		/* NOTE: you cannot send (long long) type directly */
		clientreq.length_h=totaldatasize/1000000;
		clientreq.length_l=totaldatasize%1000000;
		break;

            case RT_SABUL_VERSION:
		//  get version here 
                fprintf(stderr, "Version: %s %f\n", package, version);
                exit(EXIT_FAILURE);
                break;
	    case RT_PACKET_SIZE:
		sblop.m_iPayloadSize=sblop.m_iPktSize-sizeof(long);
		sblop.m_iMaxLossLength=sblop.m_iPktSize/sizeof(long)-2;
		break;

	    case RT_VERBOSE:

		break;
	

            default:
                fprintf(stderr, "Unsupported option: %d\n", v);
        }

    }

    if(sblop.m_iPktSize>1472)fprintf(stderr, "WARNING: Please make sure you can use jumbo packet size!!\n");

    if((mode!=0)&&(mode!=1))
    {
	fprintf(stderr, "You must specify the either server mode or client mode by \"-s\" or \"-c\" \n");
        poptPrintHelp(optCon, stderr, 0);
	exit(-1);
    }


    if ( v < -1 )
    {
        fprintf(stderr, "error: %d\n", v);
        fprintf(stderr, "%s: %s\n",
        poptBadOption(optCon, POPT_BADOPTION_NOALIAS),
        poptStrerror(v)); 
    }



   if(mode==0)   /* run as server mode */
   {

	CSabulSender* sender = new CSabulSender;	/* initialize the sender */
	//sender->setOpt(sblop);  /*  more work here  */
	sender->setOpt(SBL_VERBOSE, &sblop.m_iVerbose, sizeof(int));
	sender->setRate(rate_init, rate_l_limit, rate_u_limit);

	
	try
	{
		if(sabulport==-1)
			sabulport=sender->open();
		else
			sabulport=sender->open(sabulport);
	}
	catch(...)
	{
		fprintf(stderr, "Failed to open New Sabul Server. Program Aborted.\n");
		return -1;
	}
	if(sabulport!=7000)
	{
		fprintf(stderr,"The real port number that SABUL has been bound is %d.\n",sabulport);
		fprintf(stderr,"Please use -p to specify the real port number on your client side.\n");
	}

	try
	{
		sender->listen();
	}
	catch(...)
	{
		fprintf(stderr, "Failed to get new connection request. Program Aborted.\n");
		return -1;
	}


	/***************  open a socket to accept connection request *************/

	unsigned int iSinSize;

	serveraddr.sin_family	=	AF_INET;
	serveraddr.sin_port	=	htons(cli_port);	
	serveraddr.sin_addr.s_addr=	INADDR_ANY;		//let server automatically fill with server IP
	memset(&(serveraddr.sin_zero),'\0',8);

	if((mysock0=socket(AF_INET, SOCK_STREAM, 0))==-1)
	{
	      perror("socket");
	      exit(-1);
	}

	// bind: 0 success, -1 failure
	if(bind(mysock0,  (struct sockaddr*)&serveraddr , sizeof(struct sockaddr)) < 0)
	{
	      perror("bind");
	      exit(-1);
	}

	//listen: 0 success, -1 failure
	if (listen(mysock0, 1) < 0)    /* listen to 5 request at a time only  */
	{
	      perror("listen");
	      exit(-1);
	}

		   
	iSinSize = sizeof(struct sockaddr_in);
	if((mysock=accept(mysock0,(struct sockaddr*)&clientaddr, &iSinSize))==-1)
	{
	      perror("accept");
	}

	close(mysock0);		/* close the listen socket immediately */

	read(mysock, (char*) &clientreq , sizeof(reqinfo));
	ntohl(clientreq.length_h);
	ntohl(clientreq.length_l);	/* covert network order to host order */
	switch(clientreq.type)
	{
		case USING_LENGTH:
			totaldatasize=(long long) clientreq.length_h*1000000+clientreq.length_l;
			break;

		case USING_FILE:	/* request a file to be sent */
			struct stat statbuf;
			if(0==stat(clientreq.filename, &statbuf))	/* get the file size*/
			{
				clientreq.length_h=statbuf.st_size/1000000;
				clientreq.length_l=statbuf.st_size%1000000;
				totaldatasize=(long long)statbuf.st_size;
			}
			else	/* fail to get the file */
			{
				clientreq.type=NO_SUCH_FILE;
			}
			break;

		default:
			break;
	}


	htonl(clientreq.length_h);
	htonl(clientreq.length_l);
	write(mysock,(char *) &clientreq, sizeof(reqinfo));

	if(clientreq.type==NO_SUCH_FILE)
	{
		fprintf(stderr, "The specified file can not be openned, transaction ends. \n");
	}
	else
	{

		if(clientreq.type==USING_FILE)	/* open file if required */
		{
				fd = open(clientreq.filename, O_RDONLY);
		}


		gettimeofday(&time1, 0);	/* get the starting time here */

		/* loop the read data block by block and send them */
		for ( long long dataleft=totaldatasize; dataleft>0 ; )
		{
			int realsize = (dataleft>blocksize)?blocksize:dataleft;
			data = new char[realsize];

			if(clientreq.type==USING_FILE)
			{
				read(fd, data, realsize);
			}

			dataleft -= realsize;

			try
			{
				while (!(sender->send(data, realsize))) usleep(10);
				if (block) delete [] data;
			}
			catch(...)
			{
				fprintf(stderr, "Connection broken ...\n");
				return 0;
			}
		}


		while (int i=sender->getCurrBufSize()) 
		{	
			usleep(10);	/* if data still be sending, wait a while */
		}

		gettimeofday(&time2, 0);	/* get the end time */
		fprintf(stderr, "\n\n\nStatistics data for this transaction:\n");
		double timeused= time2.tv_sec - time1.tv_sec + (time2.tv_usec - time1.tv_usec) / 1000000.0;
		double throughput = 8*totaldatasize/timeused/1000000;
		fprintf(stderr, "Length of data: \t\t%i(M) and %d(Bytes)\n",(int)(totaldatasize/1000000),(int)(totaldatasize%1000000));
		fprintf(stderr, "Time used: \t\t\t%i(seconds)\n", (int) timeused); 
		fprintf(stderr, "Throughput: \t\t\t%5.2f(Mb/s)\n", throughput);

		if(clientreq.type==USING_FILE)
		{
			close(fd);
		}


	}

	sleep(1);	/* let the client close first */
	sender->close();	/* close SABUL */
	delete sender;		/* clear class environment */
	close(mysock);		/* close cmdline socket */



   }
   else    /********  run as client mode **********/
   {

	/**********  Open Sabul as Client    ************/
   	CSabulRecver* recver = new CSabulRecver;
	//recver->setOpt(sblop);	/********* MORE WORK HERE  **********/
	recver->setOpt(SBL_VERBOSE, &sblop.m_iVerbose, sizeof(int));
	recver->setRate(rate_init, rate_l_limit, rate_u_limit); 

	try
	{
		if(sabulport==-1)	/* if user does not provide port no for SABUL */
		{
			sabulport=7000;	/* should use DEFAULT_PORT  */

		}
		fprintf(stderr, "Connecting to %s on port %d ....\n", server_ipaddr, sabulport);
		recver->open(server_ipaddr, sabulport);
	}
	catch(...)
	{
		fprintf(stderr, "Failed to open new Sabul connection. Program Aborted.\n");
		return 0;
	}


	sleep(1);	/* wait for the server side get ready  */
	/*********   open connection to exchange transaction information  *********/
	struct hostent *sver;

	if((sver=gethostbyname(server_ipaddr))==NULL)
	{
	      perror("gethostbyname");
	      exit(-1);
	}

	if((mysock=socket(AF_INET, SOCK_STREAM, 0))==-1)
	{
	      perror("unable to open socket");
	      exit(-1);
	}

	serveraddr.sin_family	=AF_INET;
	serveraddr.sin_port	=htons(cli_port);
	serveraddr.sin_addr	=*((struct in_addr *)sver->h_addr);
	memset(&(serveraddr.sin_zero),'\0', 8);

	// connect()--- -1 failure, 0 success
	if(connect(mysock, (struct sockaddr*)&serveraddr, sizeof(struct sockaddr))==-1)
	{
	      perror("connect to CLI port ");
	}

	htonl(clientreq.length_h);
	htonl(clientreq.length_l);
        write(mysock,(char *) &clientreq, sizeof(reqinfo));    /***** exchange information here  *****/
	read(mysock, (char*) &clientreq , sizeof(reqinfo));
	ntohl(clientreq.length_h);
	ntohl(clientreq.length_l);

	totaldatasize=(long long) clientreq.length_h*1000000+clientreq.length_l;

	if(clientreq.type==NO_SUCH_FILE)
	{

		fprintf(stderr, "File can not be openned on the server side, transaction ends. \n");

	}
	else
	{
		
		if(clientreq.type==USING_FILE)
		{
			if(outfilename!=NULL)	/* output filename specified */
			{
				fd = open(outfilename, O_WRONLY|O_CREAT, S_IRWXU);
				fprintf(stderr, "Writing to file %s ....\n", outfilename);

			}
			else		/* open the output file with the name of requested filename */
			{
				fd = open(clientreq.filename, O_WRONLY|O_CREAT, S_IRWXU);
				fprintf(stderr, "Writing to file %s ....\n", clientreq.filename);

			}
		}

		data = new char[blocksize];	/* only new this memory once */ 
		gettimeofday(&time1, 0);

		for (long long dataleft=totaldatasize; dataleft >0; )
		{
			long realsize= (dataleft>blocksize)?blocksize:dataleft;
			dataleft -= realsize;
			recver->recv(data, realsize);
			if(clientreq.type==USING_FILE)
			{
				write(fd, data, realsize);
			}

		}
		gettimeofday(&time2, 0);
		delete [] data;


		fprintf(stderr, "\n\n\nStatistics data for this transaction:\n");
		double timeused= time2.tv_sec - time1.tv_sec + (time2.tv_usec - time1.tv_usec) / 1000000.0;
		double throughput = 8*totaldatasize/timeused/1000000;
		fprintf(stderr, "Length of data: \t\t%i(M) and %d(Bytes)\n",(int)(totaldatasize/1000000),(int)(totaldatasize%1000000));
		fprintf(stderr, "Time used: \t\t\t%i(seconds)\n", (int) timeused); 
		fprintf(stderr, "Throughput: \t\t\t%5.2f(Mb/s)\n", throughput);
	}

	if(clientreq.type==USING_FILE)
	{
		close(fd);
	}
	close(mysock);
	recver->close();
	delete recver;

   }


   poptFreeContext(optCon);	/* must call this !!!, release memory being used */
   return 1;
}





