#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <assert.h>

#ifndef uint16_t
typedef unsigned short uint16_t;
#endif

#define SERVER  "/tmp/serversocket"
#define MAXMSG  1500

int
main (void)
{
  extern int make_named_socket (uint16_t port);
  extern int make_socket (const char *hostname, uint16_t port);
  int sock;
  char message[MAXMSG];
  struct sockaddr_un name;
  size_t size;
  int nbytes;
#define MAXBN 1000000
  int prevblocknumber, newblocknumber;
  char *bn;  /* array of block numbers */

  /* Remove the filename first, it's ok if the call fails */
  unlink (SERVER);

  /* Allocate a byte (inited to zero) for each allowed block number. */
  assert( (bn = calloc(MAXBN, sizeof(char))) );

  /* Make the socket, then loop endlessly. */
#if LOCAL
  sock = make_named_socket (SERVER);
#else
  sock = make_socket ("*", 5555);
#endif
  {
    /* Query the size of socket receive buffer. */
    size_t siz;
    socklen_t sizsiz;

    sizsiz = sizeof(siz);
    siz = 5000000;
    assert( 0 == setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &siz, sizsiz) );
    assert( 0 == getsockopt(sock, SOL_SOCKET, SO_RCVBUF, &siz, &sizsiz) );
    assert( sizsiz == sizeof(siz) );
    fprintf(stderr, "Socket receive buffer %d bytes.\n", siz);
  }
  while (1)
    {
      /* Wait for a datagram. */
      size = sizeof (name);
      nbytes = recvfrom (sock, message, MAXMSG, 0,
                         (struct sockaddr *) & name, &size);
      if (nbytes < 0)
        {
          perror ("recvfrom (server)");
          exit (EXIT_FAILURE);
        }

      /* Process the message. */
      sscanf(message, "%d", &newblocknumber);
      if ((newblocknumber < 0) || (newblocknumber >= MAXBN)) {
        fprintf (stderr, "Block number %d out of range, terminating.\n",
                 newblocknumber);
        break;
      }
      prevblocknumber = newblocknumber;
      bn[newblocknumber]++;

#if 0
      /* Give a diagnostic message. */
      fprintf (stderr, "Server: got message: %s\n", message);
      /* xxx: Can examine 'name.xyz' to find out where
         the message came from. */

      /* Bounce the message back to the sender. */
      nbytes = sendto (sock, message, nbytes, 0,
                       (struct sockaddr *) & name, size);
      if (nbytes < 0)
        {
          perror ("sendto (server)");
          exit (EXIT_FAILURE);
        }
#endif
    }  /* while forever */

  /* Report missing blocks. */
  {
    int missing = 0;
    int i;

    for (i=0; i<MAXBN; i++) {
      if (bn[i]==0) {
        missing++;
      }
    }  /* for */
    fprintf (stderr, "Server: missed %d blocks.\n", missing);
  }
  free(bn);

}  /* main */
