
/* File2net.c 
 * Call me File2net.  I establish a socket connection to a designated 
 * Mark-5 machine and send a file or part of a file to it.  First (!) 
 * command net2disc or net2out in the Mark-5 machine, then:
 *   File2net <machine name> [ <file name> [ <startbyte> [ <endbyte> ] ] ] 
 * where <machine name> is the required name of the Mark-5 machine to 
 * send to, <file name> defaults to "save.data", <startbyte> defaults 
 * to 0, and <endbyte> defaults to the length of the file.  Progress 
 * can be monitored on the Mark-5 machine.  I end when the prescribed 
 * transfer is done, or end me with ^C.  Note that m5data tcp needs to 
 * be in /etc/services on this machine.  Copy from the Mark-5 machine 
 * if need be.  This version of File2net has been diddled so that it 
 * compiles with gcc or g++ on Linux or with gcc or cc -Ae on HP-UX. 
 * This version now supports file sizes larger than 2 Gbytes. 
 * Revised:  2002 December 2, JAB */ 
#define _LARGEFILE_SOURCE /* Large File Support (LFS) */ 
#define _FILE_OFFSET_BITS 64 /* With function replacement to 64 bits */ 
#include <stdio.h>
#include <string.h> /* For strrchr() */
#include <stdlib.h> /* For atoi() */
#include <sys/types.h> /* For open(), socket(), connect(), etc. */
#include <sys/stat.h> /* For open(), fchmod(), etc. */
#include <unistd.h> /* For fcntl(), fstat(), etc. */
#include <fcntl.h> /* For open() and fcntl() */
#include <sys/socket.h> /* For socket(), connect(), etc. */
#include <netinet/in.h> /* For socket() with PF_INET */
#include <netdb.h> /* For getservbyname() */ 
extern void herror(const char * s); /* Needed on HP-UX */ 
extern int h_errno; /* For herror() */ 
#define MIN(a,b) ((a) < (b) ? (a) : (b)) 
char * me; /* My name */
int msglev = 1; /* Debug level, default */

int main(int argc, char * argv[]) { /* File2net */
  char * machine; /* (No default) */ 
  char * filename = "save.data";
  long long startbyte = 0; 
  long long endbyte = -1; 
  long long ix; 
  int port, sock, i, k, len; 
  FILE * file; 
  struct servent * set; /* From getservbyname() */
  struct sockaddr_in socaddin;  /* For connect() socket info */ 
  struct hostent * hostinfo; /* From gethostbyname() */ 
  struct stat stats; /* For fstat() */ 
  unsigned char * uc; /* To debug print IP address */
  unsigned long rbuff[131072]; /* For data */

  /* *** Initialize *** */
  me = (me = strrchr(argv[0], '/')) == NULL ? argv[0] : me+1; /* My name */
  if (argc > 1) /* Machine name on command line? */
    machine = argv[1]; /* Yes */
  else { /* No machine name, ... */ 
    (void) fprintf(stderr, /* This is an error */ 
        "%s ERROR: \007 Machine name is required.  Usage: \n" 
        "  %s <machinename> [ <filename> [ <startbyte> [ <endbyte> ] ] ] \n", 
        me, me); 
    return(-1); /* Error */ } 
  if (argc > 2 && argv[2][0] != '.') /* File name on command line? */ 
    filename = argv[2]; /* Yes */ 
  if (argc > 3 && argv[3][0] != '.') /* Start byte on command line? */ 
    (void) sscanf(argv[3], "%lld", &startbyte); /* Yes */ 
        /* startbyte = atoll(argv[3]); */ 
        /* (Because HP-UX has no atoll()) */ 
  if (argc > 4 && argv[4][0] != '.') /* End byte on command line? */ 
    (void) sscanf(argv[4], "%lld", &endbyte); /* Yes */ 
    /* endbyte = atoll(argv[4]); */ 
  if (argc > 5) /* Debug level on command line? */ 
    msglev = atoi(argv[5]); /* Yes */ 
  /* ** Open input file ** */ 
  if ((port = open(filename, O_RDONLY)) < 0) { /* OK? */
    (void) fprintf(stderr, "%s ERROR: \007 Can't open() file ", me); /* Nope */
    perror(filename);
    return(-2); /* Error */ } 
  /* * Open this input file also as a stream * */
  if ((file = fdopen(port, "r")) == NULL) { /* OK? */
    (void) fprintf(stderr, "%s ERROR: \007 Can't fdopen() file ", me); /* No */
    perror(filename);
    return(-3); /* Error */ }
  /* ** Create and configure a socket to send data to ** */
  if ((sock = socket(PF_INET, SOCK_STREAM, 0)) < 0) { /* OK? */
    (void) fprintf(stderr, /* Nope */
        "%s ERROR: \007 socket() returned %d ", me, sock);
    perror("error");
    return(-4); /* Error */ }
  /* * Get service number for socket m5data * */
  if ((set = getservbyname("m5data", "tcp")) == NULL) { /* OK? */
    (void) fprintf(stderr, /* Nope */
        "%s ERROR: \007 m5data tcp not found in /etc/services \n", me);
    return(-5); /* Error */ } 
  socaddin.sin_port = set->s_port; /* Port service's number */ 
  if (msglev < 1) /* Debuggery? */
    /* Yes.  For Intelendian, we need to swap bytes in set->s_port */
    (void) fprintf(stderr, "%s DEBUG:  m5data tcp port is %d \n", 
        me, ntohs(socaddin.sin_port));
  socaddin.sin_family = PF_INET; /* To agree with socket() above */ 
  /* * Find IP address of machine to connect to * */
  if ((hostinfo = gethostbyname(machine)) == NULL) { /* Get IP, OK? */
    (void) fprintf(stderr, /* Nope */
        "%s ERROR: \007 gethostbyname() on %s returned NULL ",
        me, machine);
    herror("error"); /* Error */
    switch (h_errno) { /* Which error? */
      case HOST_NOT_FOUND :
        (void) fprintf(stderr, "%s ERROR:  host %s not found \n",
            me, machine);
        break;
      case TRY_AGAIN :
        (void) fprintf(stderr,
            "%s ERROR:  no response, try again later \n", me);
        break;
      case NO_RECOVERY :
        (void) fprintf(stderr,
            "%s ERROR:  unknown error, not recoverable \n", me);
        break;
      case NO_ADDRESS : /* = NO_DATA */
        (void) fprintf(stderr,
            "%s ERROR:  No Internet address available \n", me);
      } /* End of switch */
    return(-6); /* Error */
    } /* End of if hostinfo NULL */
  if (hostinfo->h_addr_list[0] == NULL) { /* First IP address OK? */
    (void) fprintf(stderr, /* Nope */
        "%s ERROR: \007 gethostbyname() on %s "
        "returned NULL IP address \n", me, machine);
    return(-7); /* Error */ }
  if (msglev < 1) { /* Debuggery? */
    uc = (unsigned char *) hostinfo->h_addr_list[0]; /* Yes */
    (void) fprintf(stderr, 
        "%s DEBUG:  IP address of %s is [", me, machine);
    for (i = 0; i < hostinfo->h_length; i++) {
      if (i > 0)
        (void) printf(".");
      (void) printf("%u", uc[i]); }
    (void) printf("] \n"); }
  socaddin.sin_addr.s_addr = *((unsigned long *) hostinfo->h_addr_list[0]); 
      /* Use first address */
  /* * Connect this socket to service on machine * */ 
  if (msglev < 1) /* Debuggery? */
    (void) fprintf(stderr, /* Yes */
        "%s DEBUG:  Trying to connect() \n", me);
  if (connect(sock, (const struct sockaddr *) &socaddin,
      sizeof(struct sockaddr_in)) < 0) { /* Connect, errors? */
    (void) fprintf(stderr, /* Yes */
        "%s ERROR: \007 connect() returned ", me);
    perror("error");
    return(-8); /* Error */ }
  if (msglev < 1) /* Debuggery? */
    (void) fprintf(stderr, /* Yes */
        "%s DEBUG:  Got a connect() on sock %d \n", me, sock); 
  /* *** Read data from the file and write to the socket (net) *** */ 
  ix = startbyte; /* Initialize */ 
  if (endbyte <= startbyte) { /* Whole file? */ 
    if ((k = fstat(port, &stats)) != 0) { /* Yes, get stats, OK? */ 
      (void) fprintf(stderr, /* Nope */ 
          "%s ERROR: \007 fstat() returned %d on ", me, k);
      perror(filename);
      return(-9); /* Error */ } 
    endbyte = stats.st_size - startbyte; /* Revised for whole file */ } 
  while (1) { /* Loop until break */ 
    /* * Read data from file * */ 
    len = MIN(sizeof(rbuff), endbyte - ix); /* Size of this chunk */ 
    k = fread((void *) rbuff, len, 1, file); /* Read a chunk */ 
    /* Check for errors */ 
    if (k != 1) { /* OK? */ 
      (void) fprintf(stderr, /* Nope */ 
          "%s ERROR: \007 fread() returned %d on file ", me, k); 
      perror(filename); 
      return(-10); /* Error */ } 
    /* * OK, send data to the socket * */ 
    k = send(sock, (void *) rbuff, len, 0); 
    if (k < 0) { /* Errors? */ 
      (void) fprintf(stderr, /* Yes */ 
          "%s ERROR: \007 Send() on socket %d returned %d ", me, sock, k); 
      perror("error"); 
      return(-11); /* Error */ } 
    /* Else OK */ 
    ix += len; /* Bump counter */ 
    if (ix >= endbyte) { /* Done? */ 
      if (msglev < 0) /* Yes.  Debuggery? */
        (void) fprintf(stderr, " \n"); /* Yes */ 
      if (msglev < 1) /* More debuggery? */ 
        (void) fprintf(stderr, /* Yes */ 
            "%s DEBUG:  ix %Ld, endbyte %Ld; done \n", me, ix, endbyte); 
      break; /* Out of while loop until break */ } 
    if (msglev < 0) /* Debuggery? */
      (void) fprintf(stderr, "#"); /* Yes */
    } /* End of while loop until break */ 
  (void) fprintf(stderr, "%s:  The End \n", me);
  return(0); /* Normal end */
  } /* End of main = File2net */

